#include "DOMSaver.hpp"

std::wostream& operator<<( std::wostream& i_ostream, const DOM::Node<std::wstring>& i_toWrite);
std::wostream& operator<<( std::wostream& i_ostream, const DOM::Node<std::string>& i_toWrite);
std::ostream& operator<<( std::ostream& i_ostream, const DOM::Node<std::string>& i_toWrite);

std::ostream& operator<<( std::ostream& i_ostream, const DOM::Node<std::wstring>& i_toWrite);
{
    switch ( i_toWrite.getNodeType())
    {
		case DOM::Node<StringT>::TEXT_NODE:
        {
			i_ostream << i_toWrite.getNodeValue();
            break;
        }

		case DOM::Node<StringT>::PROCESSING_INSTRUCTION_NODE :
        {
			i_ostream << "<?"  << i_toWrite.getNodeName() << ' ' << i_toWrite.getNodeValue() << "?>";
            break;
        }


		case DOM::Node<StringT>::DOCUMENT_NODE :
        {
			i_ostream << "<?xml version=\"1.0\" encoding=\"utf-8\"?>" << std::endl;
            for ( DOM::Node<StringT> child = i_toWrite.getFirstChild(); child != 0; child = child.getNextSibling())
            {
                i_ostream << child;
                // add linefeed in requested output encoding
				i_ostream << std::endl;
            }
            break;
        }


		case DOM::Node<StringT>::ELEMENT_NODE :
        {
            // The name has to be representable without any escapes
            i_ostream << '<' << i_toWrite.getNodeName();

            // Output the element start tag.

            // Output any attributes on this element
			DOM::NamedNodeMap<StringT> attributes = i_toWrite.getAttributes();
            for ( unsigned int i = 0; i < attributes.getLength(); i++)
            {
				DOM::Node<StringT> attribute = attributes.item(i);
				i_ostream  << ' ' << attribute.getNodeName() << "=\"" << attribute.getNodeValue() << '\"';
            }

            //
            //  Test for the presence of children, which includes both
            //  text content and nested elements.
            //
			DOM::Node<StringT> child = i_toWrite.getFirstChild();
            if (child != 0)
            {
                // There are children. Close start-tag, and output children.
                // No escapes are legal here
				i_ostream << '>';

                for (; child != 0; child = child.getNextSibling())
                {
                    i_ostream << child;
                }

                //
                // Done with children.  Output the end tag.
                //
				i_ostream << "</" << i_toWrite.getNodeName() << '>';
            }
            else
            {
                //
                //  There were no children. Output the short form close of
                //  the element start tag, making it an empty-element tag.
                //
				i_ostream << "/>";
            }
            break;
        }


		case DOM::Node<StringT>::ENTITY_REFERENCE_NODE:
        {
			DOM::Node<StringT> child;
            for ( child = i_toWrite.getFirstChild(); child != 0; child = child.getNextSibling())
            {
                i_ostream << child;
            }
            break;
        }


		case DOM::Node<StringT>::CDATA_SECTION_NODE:
        {
            i_ostream << "<![CDATA[" << i_toWrite.getNodeValue() << "]]>";
            break;
        }


		case DOM::Node<StringT>::COMMENT_NODE:
        {
			i_ostream << "<!--" << i_toWrite.getNodeValue() << "-->";
            break;
        }

		/*
		case DOM::Node<StringT>::DOCUMENT_TYPE_NODE:
        {
			const DOM::DocumentType<StringT> doctype( static_cast<DOM::DocumentType<StringT> >(i_toWrite));

			i_ostream << "<!DOCTYPE " << doctype.getNodeName() << " PUBLIC \"" << 
				doctype.getPublicId() << "\" SYSTEM \"" <<
				doctype.getSystemId() << '[' << doctype.getInternalSubset() << "]>";

			break;
        }
		*/

		case DOM::Node<StringT>::ENTITY_NODE:
        {
            i_ostream << "<!ENTITY " << i_toWrite.getNodeName();

			const DOM::Entity<StringT> entity( static_cast<DOM::Entity<StringT> >(i_toWrite));
            i_ostream << "PUBLIC \"" << entity.getPublicId() << '\"';
			i_ostream << "SYSTEM \"" << entity.getSystemId() << '\"';
			i_ostream << "NDATA \"" << entity.getNotationName() << '\"';

			i_ostream << '>' << std::endl;

            break;
        }



        default:
			std::cerr << "Unrecognized node type = " << (long)i_toWrite.getNodeType() << std::endl;
    }
    return i_ostream;
}
