#include "XMLDeSerializer.hpp"
//#include "XMLUtils.hpp"

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
XERCES_CPP_NAMESPACE_USE

#include <exception>
#include <sstream>
#include <iostream>

template<class t_type>
bool DeSerializePOD( t_type& o_out, std::wstringstream& i_wstrstream, const PSF::IChecker& i_checker, size_t& io_nItemsDeSerialized)
{
	i_wstrstream >> std::boolalpha >> o_out;
	if ( !i_wstrstream.eof() && !i_wstrstream.good())
	{
		i_checker.SematicError();
	}
	if ( i_wstrstream.good()) io_nItemsDeSerialized++;
	return i_wstrstream.good();
}

XERCES_CPP_NAMESPACE::DOMElement* FindElement( const XERCES_CPP_NAMESPACE::DOMElement* i_pdomElementParent, const std::wstring& i_wstrName)
{
	if ( !i_pdomElementParent) return 0;
	XERCES_CPP_NAMESPACE::DOMNode* pdomNode;
	for ( pdomNode = i_pdomElementParent->getFirstChild(); pdomNode != 0; pdomNode = pdomNode->getNextSibling())
	{
		if ( XERCES_CPP_NAMESPACE::DOMNode::ELEMENT_NODE != pdomNode->getNodeType() || i_wstrName != pdomNode->getNodeName()) continue;
		
		return (XERCES_CPP_NAMESPACE::DOMElement*)pdomNode;
	}
	return 0;
}

XERCES_CPP_NAMESPACE::DOMElement* NextElement( const XERCES_CPP_NAMESPACE::DOMElement* i_pdomElement)
{
	if ( !i_pdomElement) return 0;
	XERCES_CPP_NAMESPACE::DOMNode* pdomNode = i_pdomElement->getNextSibling();
	while ( pdomNode && XERCES_CPP_NAMESPACE::DOMNode::ELEMENT_NODE != pdomNode->getNodeType())
	{
		pdomNode = pdomNode->getNextSibling();
	}

	return (XERCES_CPP_NAMESPACE::DOMElement*)pdomNode;
}

//////////////////////////////////////////////////////////////////////////////////////
// CXMLDeSerializerBase : Implmentation
//////////////////////////////////////////////////////////////////////////////////////

PSF::CXMLDeSerializerBase::CXMLDeSerializerBase( DOMElement* i_pdomElementRoot, 
												 bool i_bSearch,
												 const SerializerTypeMap& i_mapType,
												 const IChecker& i_checker) :
	m_pdomElementRoot( i_pdomElementRoot),
	m_pdomElementCurrent( 0),
	m_bSearch( i_bSearch),
	m_mapType( i_mapType),
	m_checker( i_checker),
	m_nItemsDeSerialized( 0),
	m_nRangeSize( -1),
	m_wstrstream( std::ios_base::in)
{
	const wchar_t* pszSize = m_pdomElementRoot->getAttribute( L"size");
	if ( pszSize)
	{
		std::wstringstream wstrstream( pszSize);
		wstrstream >> m_nRangeSize;
	}

	DOMNode* pdomNode = m_pdomElementRoot->getFirstChild();
	while ( pdomNode && DOMNode::ELEMENT_NODE != pdomNode->getNodeType())
	{
		pdomNode = pdomNode->getNextSibling();
	}

	m_pdomElementCurrent = (DOMElement*)pdomNode;

	std::wstring wstrText;
	for ( pdomNode = m_pdomElementRoot->getFirstChild(); pdomNode != 0; pdomNode = pdomNode->getNextSibling())
	{
		if ( pdomNode->getNodeType() == XERCES_CPP_NAMESPACE::DOMNode::TEXT_NODE)
		{
			wstrText += pdomNode->getNodeValue();
		}
	}

	m_wstrstream.str( wstrText);
}


PSF::CXMLDeSerializerBase::~CXMLDeSerializerBase()
{
	if ( m_nRangeSize != -1)
	{
		if ( m_nItemsDeSerialized != m_nRangeSize)
		{
			m_checker.RangeSizeError();
			// emit range error
		}
	}
	// todo : assert member are null
}

XERCES_CPP_NAMESPACE::DOMElement*  PSF::CXMLDeSerializerBase::UpdateCurrentElement( const std::wstring& i_wstrName)
{
	while ( m_pdomElementCurrent && i_wstrName != m_pdomElementCurrent->getTagName())
	{
		m_pdomElementCurrent = NextElement( m_pdomElementCurrent);
	}

	if ( !m_pdomElementCurrent && m_bSearch)
	{
		m_pdomElementCurrent = FindElement( m_pdomElementRoot, i_wstrName);
	}

	DOMElement* pdomElement = m_pdomElementCurrent;
	m_pdomElementCurrent = NextElement( m_pdomElementCurrent);
	return pdomElement;
}

bool PSF::CXMLDeSerializerBase::DeSerialize( char& o_out)
{
	wchar_t wch;
	m_wstrstream >> std::noskipws;
	bool bResult = DeSerializePOD( wch, m_wstrstream, m_checker, m_nItemsDeSerialized);
	m_wstrstream >> std::skipws;
	if ( bResult) o_out = wch;
	return bResult;
}

bool PSF::CXMLDeSerializerBase::DeSerialize( unsigned char& o_out)
{
	short n;
	bool bResult = DeSerializePOD( n, m_wstrstream, m_checker, m_nItemsDeSerialized);
	if ( bResult) o_out = n;
	return bResult;
}

bool PSF::CXMLDeSerializerBase::DeSerialize( short& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( unsigned short& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( int& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( unsigned int& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( long& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( unsigned long& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( float& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( double& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

bool PSF::CXMLDeSerializerBase::DeSerialize( wchar_t& o_out)
{
	// streams skip whitespace per default, but for strings it is important
	m_wstrstream >> std::noskipws;
	bool bResult = DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
	m_wstrstream >> std::skipws;
	return bResult;
}

bool PSF::CXMLDeSerializerBase::DeSerialize( bool& o_out)
{
	return DeSerializePOD( o_out, m_wstrstream, m_checker, m_nItemsDeSerialized);
}

std::auto_ptr<PSF::IDeSerializer> PSF::CXMLDeSerializerBase::DeSerializeCompound( const std::wstring& i_wstrName, 
																				  const Loki::TypeInfo& i_type)
{
	DOMElement* pdomElement = UpdateCurrentElement( i_wstrName);

	if ( !pdomElement) return std::auto_ptr<IDeSerializer>( 0);

	CheckType( pdomElement->getAttribute( L"type"), i_type);
	m_nItemsDeSerialized++;
	return std::auto_ptr<IDeSerializer>( new CXMLDeSerializerBase( pdomElement, true, m_mapType, m_checker));
}

std::auto_ptr<PSF::IDeSerializer> PSF::CXMLDeSerializerBase::DeSerializeRange( const std::wstring& i_wstrName, 
																			   const Loki::TypeInfo& i_type)
{
	DOMElement* pdomElement = UpdateCurrentElement( i_wstrName);

	if ( !pdomElement) return std::auto_ptr<IDeSerializer>( 0);

	CheckType( pdomElement->getAttribute( L"type"), i_type);
	m_nItemsDeSerialized++;
	return std::auto_ptr<IDeSerializer>( new CXMLDeSerializerBase( pdomElement, false, m_mapType, m_checker));
}

void PSF::CXMLDeSerializerBase::CheckType( const std::wstring& i_wstrType, const Loki::TypeInfo& i_type)
{
	if ( i_wstrType.empty())
	{
		// No type info
		m_checker.TypeMissing();
		return;
	}

	SerializerTypeMap::const_iterator iterType = m_mapType.find( i_type);
	if ( m_mapType.end() == iterType)
	{
		// Unknown type
		m_checker.TypeUnknown();
		return;
	}

	if ( iterType->second != i_wstrType)
	{
		// Type mismatch
		m_checker.TypeMismatch();
	}
}

//////////////////////////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////////////////////////

PSF::CXMLDeSerializerProxy::CXMLDeSerializerProxy( const std::wstring& i_wstrFilename,
										 const SerializerTypeMap& i_mapType,
										 const IChecker& i_checker) :
	m_pdomImpl( 0),
	m_pdomBuilder( 0),
	m_pdomDoc( 0),
	m_pBase( 0)
{
	try 
	{
		XMLPlatformUtils::Initialize();
		m_pdomImpl = (DOMImplementation*)DOMImplementationRegistry::getDOMImplementation( L"Core XML 1.0 LS");
        m_pdomBuilder = m_pdomImpl->createDOMBuilder( DOMImplementationLS::MODE_SYNCHRONOUS, 0);

		m_pdomDoc = m_pdomBuilder->parseURI( i_wstrFilename.c_str());

		m_pBase = std::auto_ptr<CXMLDeSerializerBase>( new CXMLDeSerializerBase( m_pdomDoc->getDocumentElement(), true, i_mapType, i_checker));
	}
	catch (const DOMException& toCatch) {
		char* pszMessage = XMLString::transcode( toCatch.msg);
		throw std::exception( pszMessage);
	}
	catch (const XMLException& toCatch) 
	{
		char* pszMessage = XMLString::transcode( toCatch.getMessage());
		throw std::exception( pszMessage);
	}
}

PSF::CXMLDeSerializerProxy::~CXMLDeSerializerProxy()
{
	m_pdomBuilder->release();
    XMLPlatformUtils::Terminate();
}

PSF::CXMLDeSerializerProxy::operator PSF::IDeSerializer&()
{
	return *m_pBase;
}
